#==============================================================================
#
#  $Id: mainwin.py,v 1.22 2001/03/13 02:28:35 mike Exp $
#
"""
   This module houses the spugmail main window (the folder viewer) and all of 
   its supporting windows.
"""
#
#  Copyright (C) 1999 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#  $Log: mainwin.py,v $
#  Revision 1.22  2001/03/13 02:28:35  mike
#  Added functions to check and force a write of the current folder.
#
#  Revision 1.21  2000/06/29 00:59:10  mike
#  Minor fix: no folder name aborts when creating a new folder.
#
#  Revision 1.20  2000/06/25 21:36:10  mike
#  Added message forwarding.
#
#  Revision 1.19  2000/06/16 00:03:32  mike
#  Added the "entry menu" widget, used it for folder selection.
#
#  Revision 1.18  2000/05/28 20:43:11  mike
#  Made title show the current folder, added a hack to make it still
#  show it after the editor was launched.
#
#  Revision 1.17  2000/05/28 19:53:52  mike
#  Added the ability to save messages to the "Drafts" folder.
#
#  Revision 1.16  2000/05/25 01:27:20  mike
#  queryYesNo() did not need a "client" parm.
#
#  Revision 1.15  2000/05/19 00:28:56  mike
#  Bug fix - was writing current mailbox twice.
#
#  Revision 1.14  2000/05/19 00:04:20  mike
#  Using new curses code to switch modes before changing to a full screen app.
#
#  Revision 1.13  2000/05/11 01:09:53  mike
#  Added attachment processing.
#
#  Revision 1.12  2000/05/03 23:49:22  mike
#  Fixed bug: was editing live message index in the list viewer.
#
#  Revision 1.11  2000/04/30 21:07:22  mike
#  Added initial address book window.
#
#  Revision 1.10  2000/04/27 23:23:50  mike
#  Added "auto-move to folder", fixed bug that occurred with no date header,
#  misc cleanup.
#
#  Revision 1.9  2000/04/17 12:34:58  mike
#  Refer to custom instead of main for message instance creation.
#
#  Revision 1.8  2000/04/16 17:44:03  mike
#  Added code to draw a box around a text message; added code to create the
#  inbox if it does not exist.
#
#  Revision 1.7  2000/04/14 01:21:52  mike
#  Added folder support.
#
#  Revision 1.6  2000/04/12 00:43:54  mike
#  Added hidden entryfields and initial support for customization.
#
#  Revision 1.5  2000/04/06 01:14:18  mike
#  Added message menu items in main win to the message viewer.
#
#  Revision 1.4  2000/04/06 00:51:24  mike
#  Added "write to file" and "pipe to command" options.
#
#  Revision 1.3  2000/04/04 01:12:01  mike
#  Added delete message functionality.
#
#  Revision 1.2  2000/04/02 22:39:36  mike
#  Added the menu interface, added the "compose" option.
#
#  Revision 1.1.1.1  2000/04/02 19:09:03  mike
#  Spug Mailer and libraries.
#
#
#==============================================================================

import sys, time, os, popen2, string, copy
sys.path.insert(0, '../..')

from spug.curses.App import App
from spug.curses.KeyEvent import KeyEvent
from spug.curses.Listbox import Listbox
from spug.curses.Label import Label
from spug.curses.ModalWindow import ModalWindow
from spug.curses.SimpleTextViewer import SimpleTextViewer
from spug.curses.StatusText import StatusText
from spug.curses.Entryfield import Entryfield
from mailbx import Mailbox
from composer import Composer
from addrwin import AddrList
from attchmt import AttachmentsViewer

import main

class QueryLine(ModalWindow):
   """
      This just pops up over the status line to allow us to ask a question.
   """
   def __init__(self, x0, y0, width, height, prompt, initValue = '', 
                secret = 0
                ):
      ModalWindow.__init__(self, x0, y0, width, height)
      self.__label = Label(self, 0, 0, prompt)
      promptLen = len(prompt)
      self.__text = Entryfield(self, promptLen, 0, width - promptLen - 1,
                               200,
                               value = initValue,
                               secret = secret
                               )
   
   def processEvent(self, evt):
      if isinstance(evt, KeyEvent):
         if evt.code == 10:
            self.terminate()
            return 1
      
      return ModalWindow.processEvent(self, evt)
   
   def getText(self):
      """
         Returns the text entered into the entryfield.
      """
      return self.__text.getText()

class YesNoQueryLine(ModalWindow):
   """
      This pops up over the status line to allow us to ask a yes-no
      question.
   """
   
   def __init__(self, x0, y0, width, height, prompt):
      ModalWindow.__init__(self, x0, y0, width, height)
      self.__label = Label(self, 0, 0, prompt + ' [Y/N]')
   
   def processEvent(self, evt):
      if isinstance(evt, KeyEvent):
         if evt.code in (89, 121):
            self.isYes = 1
         elif evt.code in (78, 110):
            self.isYes = 0
         else:
            self.beep()
            return
         self.terminate()

class TextBox(ModalWindow):
   """
      Just a small scrolling text viewer.
   """

   def __init__(self, x0, y0, width, height, message):
      ModalWindow.__init__(self, x0, y0, width, height)
      self.box()
      self.refresh()
      self.text = SimpleTextViewer(self, 1, 1, width - 2, height - 2, message)
   
   def processEvent(self, evt):
   
      if ModalWindow.processEvent(self, evt):
         return 1

      # terminate on any keystroke         
      if isinstance(evt, KeyEvent):
         self.terminate()
         return 1
      
      return 0

class MessageViewer(ModalWindow):
   
   def __init__(self, x0, y0, width, height, message):
      ModalWindow.__init__(self, x0, y0, width, height)
      self.text = SimpleTextViewer(self, 0, 0, width - 1, height, message)
      self.__menu = \
         [
         (114, 'Reply to all', self.replyToAll),
         (82, 'reply to sender (Shift-R)', self.replyToMessage),
         (102, 'Forward', self.forwardMessage),
         (100, 'toggle Delete (also Del)', self.toggleDelete),
         (109, 'Move to folder', self.moveToFolder),
         (77, 'auto-move to folder (Shift-M)', self.autoMove),
         (119, 'Write to file', mainWin.writeToFile),
         (124, 'pipe to command ("|")', self.pipeToCommand),
         (118, 'View attachments', self.viewAttachments),
         (113, 'Quit', self.quit),
         ]         
   
   def processEvent(self, evt):
      if isinstance(evt, KeyEvent):
         if evt.code == self.KEY_LEFT:
            self.quit()
            return 1
         elif evt.code == 24:            # Ctrl-X
            main.menu( self.__menu,
                      self
                      )
            return 1
      
      return ModalWindow.processEvent(self, evt)

   def replyToAll(self):
      mainWin.replyToMessage(1, self)
   
   def replyToMessage(self):
      mainWin.replyToMessage(0, self)

   def forwardMessage(self):
      mainWin.forwardMessage(self)

   def toggleDelete(self):
      mainWin.toggleDelete(0)
   
   def pipeToCommand(self):
      mainWin.pipeToCommand(self)
   
   def viewAttachments(self):
      mainWin.viewAttachments(self)
   
   def moveToFolder(self):
      mainWin.moveToFolder(0, self)

   def autoMove(self):
      mainWin.autoMove(0)
   
   def quit(self):
      del self.__menu
      self.terminate()

class MailListbox(Listbox):
   
   def __init__(self, parent, x0, y0, width, height, items):
      Listbox.__init__(self, parent, x0, y0, width, height, items)
      self.x = x0
      self.y = y0
      self.width = width
      self.height = height
   
   def _itemAsString(self, item):
      try:
         dateString = \
            time.strftime('%Y-%m-%d %H:%M:%S',
                          item.date
                          )
      except ValueError:
         dateString = '<Bad Date>'
         
      return '%c %-20.20s %-36.36s %19.19s' % \
              (item.isDeleted() and 'D' or 
                item.unread and 'N' or ' ', 
               item.sender, 
               item.subject, 
               dateString
               )

   def showSelection(self, showAllHeaders = 0):
      selection = self.selection()
      if selection:
         msg = selection.getMessage()
         text = MessageViewer(self.x, self.y, self.width, self.height,
                              msg.forDisplay(showAllHeaders)
                              )
         selection.markAsRead()
         text.show(self)
         self.paint()
   
   def processEvent(self, evt):
      if isinstance(evt, KeyEvent):
         if evt.code == ord('h'):
            self.showSelection(1)
            return 1
      return  Listbox.processEvent(self, evt)
   

class MailClient(App):


   def __init__(self):
      App.__init__(self)
      height, width = self.getmaxyx()
      self.title = StatusText(self, 0, 0, width, 1,  'SpugMail - Inbox')
      
      # get the inbox
      self.inbox = self.mailbox = main.postOffice.getMailbox('Inbox')
         
      self.currentFolder = 'Inbox'
      self.status = StatusText(self, 0, height - 1, width - 1, 1, 
                               'Loading inbox...'
                               )
      self.messageList = MailListbox(self, 0, 1, width, height - 2, 
                                     list(self.mailbox.getIndex())
                                     )
      self.status.setText('')
      
      # since there are no entryfield controls in this window, we will be
      # using this menu object to build a key table as well as as a menu
      self.__menu = \
         [ 
         (103, 'Get messages', self.getMessages),
         (99, 'Compose message', self.composeMessage),
         
         (97, 'Address book', self.addressBook),
         
         (110, 'make New folder', self.createFolder),
         (115, 'Select folder', self.selectFolder),
         (107, 'checK current folder', self.checkFolder),
         (111, 'fOrce write of index', self.writeFolder),
         
         (114, 'Reply to all', self.replyToAll),
         (82, 'reply to sender (Shift-R)', self.replyToMessage),
         (102, 'Forward', self.forwardMessage),

         (100, 'toggle Delete (also Del)', self.toggleDelete),
         (109, 'Move to folder', self.moveToFolder),
         (77, 'auto-move to folder (Shift-M)', self.autoMove),
         (119, 'Write to file', self.writeToFile),
         (124, 'pipe to command ("|")', self.pipeToCommand),
         (118, 'View attachments', self.viewAttachments),
         (113, 'Quit', self.quit),
         ]

      # initialize the keycode table
      self.keys = \
         {
         self.KEY_DC: self.toggleDelete,
         }
      
      # add menu codes to the table
      for code, text, func in self.__menu:
         self.keys[code] = func      

   def pickFolder(self, client = None):
      """
         Pop's up a menu to allow the user to select a folder.  Excludes
         the current folder from this list.  Returns the selected item,
         or *None* if nothing was selected.
         
         If /client/ is provided, should be the current client window.
         Defaults to /self/.
      """
      if not client:
         client = self
         
      # build the list of folders from the folder path.
      folders = []
      for file in main.postOffice.getMailboxNames():
         if file != self.currentFolder:
            folders.append((file, file))
      return main.entryMenu(folders, client)
      
   def getMessages(self):
      errors = []
      for loader in main.custom.getLoaders():
         errors = errors + loader.getMessages(self)
      
      self.status.setText('')

      if errors:
      
         # format an error message for display
         str = 'There were errors:\n'
         for err in errors:
            str = str + '-  %s\n' % err
         
         self.showText(str)

   def createFolder(self):
      
      # get the folder name from the user
      folder = self.queryUser('Folder Name:')
      if not folder:
         return
      
      # make sure it doesn't already exist
      if main.postOffice.hasMailbox(folder):
         self.beep()
         self.showStatus('Folder %s already exists!' % folder)
         return
      
      # create it
      main.postOffice.createMailbox(folder)

   def selectFolder(self):
      
      folder = self.pickFolder()
      if not folder:
         return
         
      self.status.setText('Loading %s...' % folder)
      self.mailbox = main.postOffice.getMailbox(folder)
      self.status.setText('')
      
      # reload the message list from the given mailbox
      self.messageList.setItems(list(self.mailbox.getIndex()))
      
      self.currentFolder = folder
      self.__setTitle()
   
   def checkFolder(self):
      if not self.queryYesNo('Check %s for inconsistencies?' % 
                              self.currentFolder
                             ):
         return
      
      self.status.setText('Checking folder %s...' % self.currentFolder)
      added, removed = self.mailbox.checkMailbox()
      if not added and not removed:
         self.status.setText('Mailbox is clean.')
      else:
         self.status.setText('Mailbox was dirty: %d messages added to index '
                              '%d messages removed.' % (added, removed)
                             )
         
         # reload the list
         self.messageList.setItems(list(self.mailbox.getIndex()))
   
   def writeFolder(self):
      self.status.setText('Writing folder %s...' % self.currentFolder)
      self.mailbox.writeIndex()
      self.status.setText('Folder %s index written.' % self.currentFolder)
   
   def __setTitle(self):
      # have to do this because touchwin doesn't seem to work - at least
      # not after we launch another app.
      self.title.setText('')
      self.title.setText('SpugMail - %s' % self.currentFolder)

   def composeMessage(self):
      Composer(0, 1, self._width, self._height - 2,
               main.custom.makeNewMessage()
               ).show(self)

   def addressBook(self):
      """
         Brings up the address book editor window.
      """
      AddrList(0, 1, self._width, self._height - 2).show(self)

   def replyToMessage(self, replyToAll = 0, owner = None):
      """
         Creates a default reply message and launches the composer with it.
         If /replyToAll/ is true, message will be created with all recipients
         in the address fields.  If /owner/ is true, the composer will be
         modal to /owner/ instead of the main window.
      """
      # by default, owner is self
      if not owner:
         owner = self
      
      # get the message and reply to it
      origMsg = self.messageList.selection().getMessage()      
      Composer(0, 1, self._width, self._height - 2, 
               main.custom.makeReplyMessage(origMsg, replyToAll)
               ).show(owner)

   def forwardMessage(self, owner = None):
      """
         Creates a "forwarded" wrapper around the message and launches the
         composer with it.
         
         parms:
            owner::
               [@spug.curses.ModalWindow or None] if provided, this is the
               window that the composer will be shown modal to.
      """
      # by default, owner is this window
      if not owner:
         owner = self
      
      # get the original message
      origMessage = self.messageList.selection().getMessage()
      
      # bring up composer
      Composer(0, 1, self._width, self._height - 1,
               main.custom.makeForwardedMessage(origMessage)
               ).show(owner)

   def replyToAll(self):
      self.replyToMessage(1)

   def toggleDelete(self, repaint = 1):
      """
         Toggles the deleted flag for the current selection.  If /repaint/ is
         true, repaints the selection.
      """
      self.messageList.selection().toggleDeleted()
      if repaint:
         self.messageList.paintSelection()

   def __moveMessageToFolder(self, msg, folder, folderName, repaint):
      """
         Private method to move a message to a folder used by both functions
         that do this.
         
         parms:
            msg::
               [@spug.mail.mailbx.MailboxMessageInfo] Message info instance
               to be moved.
            folder::
               [@spug.mail.mailbx.Mailbox] Mailbox object to move it to.
            folderName::
               [string] name of the folder
            repaint::
               [boolean] set to true if the listbox item should be repainted
               after its delete.
      """
      # pull the message data and add it to the folder selected
      msgData = msg.getMessage()
      folder.addMessage(msgData)
      
      # make sure that it is marked deleted
      if not msg.isDeleted():
         self.toggleDelete(repaint)
      
      self.showStatus('Message moved to %s and marked deleted.' % folderName)

   def moveToFolder(self, repaint = 1, client = None):
      """
         Moves the selected message to a folder specified by the user and
         marks it deleted.  If /repaint/ is true, repaints the display line.
         If /client/ is provided, should be the current client window
         (default is /self/).
      """
      # pick a folder
      folderName = self.pickFolder(client)
      if not folderName:
         return
      folder = main.postOffice.getMailbox(folderName)

      # get the message
      msg = self.messageList.selection()
      
      self.__moveMessageToFolder(msg, folder, folderName, repaint)

   def autoMove(self, repaint = 1):
      """
         Moves the selected message to a folder based on the rules identified
         in the auto-file filters.  If /repaint/ is true, repaints the
         display line.
      """
      
      # get the message
      msg = self.messageList.selection()
      msgData = msg.getMessage()
      
      # pick a folder
      for filter in main.custom.getAutoFileFilters():
         folderName = filter(msgData)
         if folderName:
            break
      else:
         self.showStatus('Message not moved: no matching rule.')
         return
      
      # get the folder
      try:
         folder = main.postOffice.getMailbox(folderName)
      except KeyError:
         main.postOffice.createMailbox(folderName)
         folder = main.postOffice.getMailbox(folderName)
      
      self.__moveMessageToFolder(msg, folder, folderName, repaint)

   def writeToFile(self):
      """
         Writes the selected message to a file.  
      """
   
      # get the file name from the user
      fname = self.queryUser('Filename:')
      if not fname:
         self.showStatus('Write Cancelled.')
         return

      # write it   
      self.messageList.selection().getMessage().writeTo(open(fname, 'w'))
      self.showStatus('Wrote to file %s.' % fname)

   def pipeToCommand(self, owner = None):
      """
         Pipes the selected message to a command.  If /owner/ is specified,
         it will be the owner of the modal text window for the command's
         output.
      """
      
      # get the command from the user, no command means cancel
      cmd = self.queryUser('Command:')
      if not cmd:
         self.showStatus('Pipe Cancelled.')
         return
      
      # write the message to the command
      msg = self.messageList.selection().getMessage()
      pout, pin, perr = popen2.popen3(cmd)
      msg.writeTo(pin)
      pin.close()
      
      # read stdout and stderr and display it, in the status line if
      # possible
      output = pout.read() + perr.read()
      if '\n' not in string.strip(output):
         self.showStatus('Output: ' + output)
      else:
         if not owner:
            owner = self
         self.showText(output, owner)      

   def viewAttachments(self, owner = None):
      """
         Brings up the attachments listbox to view the attachments for the
         selected message.
      """
      msgInfo = self.messageList.selection()
      if not msgInfo:
         self.showStatus('You must first select a message')
         return
      
      if not owner:
         owner = self
      
      msg = msgInfo.getMessage()
      AttachmentsViewer(0, 1, self._width, self._height - 2, msg).show(owner)

   def editSelection(self):
      """
         Edits the currently selected message in the composer.
         
         This method assumes that the current folder is the "Drafts" 
         folder.
      """
      selection = self.messageList.selection()
      if selection:
         msg = selection.getMessage()
         composer = Composer(0, 1, self._width, self._height - 2, 
                             copy.copy(msg))
         composer.show(self)
         
         # if the message was saved, we have to add it's info to the displayed
         # list
         if composer.messageWasSaved:
            self.messageList.append(composer.messageWasSaved)
         
   def quit(self):
      # clean up cyclic reference
      del self.__menu, self.keys
      self.terminate()
   
   def processEvent(self, evt):
      if isinstance(evt, KeyEvent):
         if evt.code == 24:            # Ctrl-X
            main.menu( self.__menu,
                      self
                      )
            return 1
         elif self.keys.has_key(evt.code):
            self.keys[evt.code]()
            return 1
         elif evt.code == 10:
            if self.currentFolder == 'Drafts':
               self.editSelection()
            else:
               self.messageList.showSelection()
            return 1
            
      
      return App.processEvent(self, evt)
   
   def showText(self, msg, owner = None):
      """
         Pops up a text message area to display a multi-line message.
         /owner/ should be the window owning the text modal message window.
      """
      # set the default value if necessary
      if owner is None:
         owner = self
         
      if self._width < 60:
         textWidth = self._width
      else:
         textWidth = 60
      xpos = (self._width - textWidth) / 2
      
      if self._height < 20:
         textHeight = self._height
      else:
         textHeight = 20
      ypos = (self._height - textHeight) / 2
      
      TextBox(xpos, ypos, textWidth, textHeight, msg).show(owner)
      self.refresh()
   
   # implements the "LoaderInterface"
   
   def showStatus(self, msg):
      self.status.setText(msg)
   
   def storeMessage(self, msg):
      # store the message in the inbox
      info = self.inbox.addMessage(msg)
      
      # store it in the listbox if we are looking at the inbox
      if self.currentFolder == 'Inbox':
         self.messageList.append(info)

   def queryUser(self, prompt, secret = 0, initValue = ''):
      queryLine = QueryLine(0, self._height - 1, self._width - 1, 1,
                            prompt,
                            secret = secret,
                            initValue = initValue
                            )
      
      queryLine.show(self.status)
      return queryLine.getText()

   def queryYesNo(self, prompt):
      """
         Prompts the user for a `yes/no` question, returns true or
         false depending on their answer.
      """
      queryLine = YesNoQueryLine(0, self._height - 1, self._width - 1, 1,
                                 prompt
                                 )
               
      queryLine.show(self.status)
      return queryLine.isYes

   def launchApp(self, cmdline, child):
      self.beforeNestedApp()
      os.system(cmdline)
      self.afterNestedApp(child)
      self.__setTitle()

mainWin = MailClient()
main.mainWin = mainWin

